#!/usr/bin/env python
"""
.. codeauthor:: Eric Hutton <huttone@colorado.edu>

.. sectionauthor:: Eric Hutton <huttone@colorado.edu>
"""

from .dupuit_percolator import GroundwaterDupuitPercolator

__all__ = ["GroundwaterDupuitPercolator"]
